local log = Winning_Log


-- NOTE:
-- A lot of this battleground data is curated based on the localized map name.
-- This is because for some reason there is not an API call that providese a
-- sane Battleground ID. The `GetBattlegroundInfo` call provides two things
-- that could be close:
--
--      battleGroundID
--      BGMapID
--
-- But the call itself is based on an index like 1-13 (which is derived from
-- the `GetNumBattlegroundTypes` call).
--
-- Other addons work around this by doing some fancy dance regarding setting
-- the zone map to the current BG and then tracking that map ID as the BG, but
-- that is pretty lame. Localized name of the BG will be sufficient until it
-- isn't.


-------------------------------------------------------------------------------
-- WinningBgRecord
--
-- Encapsulates all Blizzard calls regarding BG data.
-------------------------------------------------------------------------------

WinningBgRecord = {}
WinningBgRecord.__index = WinningBgRecord

WinningBgRecord.new = function(self, player_name)
    local self = {}
    setmetatable(self, WinningBgRecord)

    self._player_name = player_name
    local found_player = false
    for idx = 1, GetNumBattlefieldScores() do
        local name, killingBlows, honorableKills, deaths, honorGained, faction, race, class, classToken, damageDone, healingDone, bgRating, ratingChange, preMatchMMR, mmrChange, talentSpec = GetBattlefieldScore(idx)
        if name == self._player_name then
            found_player = true
            self._kbs = killingBlows
            self._hks = honorableKills
            self._deaths = deaths
            self._damage = damageDone
            self._healing = healingDone
            self._class = classToken
            self._talent_spec = GetSpecializationInfo(GetSpecialization())
            self._race = race
            self._bg_faction = faction
            self._winning_faction = GetBattlefieldWinner() or 1
            self._bg_date = time()
            self._duration = math.floor(GetBattlefieldInstanceRunTime()/1000)

            -- TODO: This is really terrible. Need to figure out a real way to
            -- get the current battleground id.
            SetMapToCurrentZone()
            self._map_id = GetCurrentMapAreaID()

            self._season = GetCurrentArenaSeason()
            self._is_rated = IsRatedBattleground()
        end
    end
    if not found_player then
        log("Player name is: " .. tostring(self._player_name))
        error("Could not find player in BG data: " .. self._player_name)
    end

    return self
end

WinningBgRecord.get_player_name = function(self)
    return self._player_name
end

WinningBgRecord.get_player_spec = function(self)
    return self._talent_spec
end

WinningBgRecord.get_player_race = function(self)
    return self._race
end

WinningBgRecord.get_date = function(self)
    return self._bg_date
end

WinningBgRecord.get_map_id = function(self)
    return self._map_id
end

WinningBgRecord.get_duration = function(self)
    return self._duration
end

WinningBgRecord.get_won = function(self)
    return self._bg_faction == self._winning_faction
end

WinningBgRecord.get_kbs = function(self)
    return self._kbs
end

WinningBgRecord.get_hks = function(self)
    return self._hks
end

WinningBgRecord.get_deaths = function(self)
    return self._deaths
end

WinningBgRecord.get_player_damage = function(self)
    return self._damage
end

WinningBgRecord.get_player_healing = function(self)
    return self._healing
end

WinningBgRecord.get_season = function(self)
    return self._season
end

WinningBgRecord.get_was_rated_match = function(self)
    return self._is_rated
end




-------------------------------------------------------------------------------
-- WinningBattlegroundRecordModel
--
-- Just a helper to document what an expected bg model looks like.
-------------------------------------------------------------------------------

WinningBattlegroundRecordModel = {
    uuid = false,
    PlayerName = false,
    PlayerSpec = false,
    PlayerRace = false,
    Type = false,
    Date = false,
    MapId = false,
    Map = false,
    DurationRaw = false,
    Duration = false,
    Won = false,
    Winner = false,
    KBs = false,
    HKs = false,
    Deaths = false,
    Damage = false,
    DamageRaw = false,
    Healing = false,
    HealingRaw = false,
    Season = false,
    Rated = false
}



-------------------------------------------------------------------------------
-- WinningDataAccessLayer
--
-- Control access to Record data.
-------------------------------------------------------------------------------

WinningDataAccessLayer = {}


WinningDataAccessLayer.init = function(self, addon)
    self.addon = addon

    self._db = addon.db.faction

    -- This is the master record of all data.
    -- WAS: self.db
    self.all_data = self._db.all_records

    -- This is a munged storage of data, fit for UI components.
    -- WAS: self.bg_db
    self.ui_data= self._db.bg_records

    -- this is a table containing bg records that are currently being shown in
    -- the UI, so that we can calculate stats from them
    -- WAS: self.ui_data_data
    self.ui_bg_record_data = self._db.bg_db_data

    -- Type contexts.
    self.CASUAL = "Casual"
    self.RATED = "Rated"

    -- Time contexts.
    self.ALL = "All"
    self.SEASON = "This Season"
    self.SIX_HOURS = "Last 6 Hours"

    -- Map contexts.
    self.MAP = "MAP"

    return self
end


WinningDataAccessLayer.get_specs = function(self)
    -- specs = {
    --  102 = "Balance",
    --  103 = "Restoration",
    --  ...
    -- }
    
    local specs = {}
    local seen_specs = {}
    -- 0 is special and means All
    table.insert(specs, 0)
    for i, bg_record in pairs(self.all_data) do
        if bg_record.PlayerSpec then
            if seen_specs[bg_record.PlayerSpec] == nil then
                log("checking spec: " .. tostring(bg_record.PlayerSpec))
                local id, name, description, icon, background, role = GetSpecializationInfoByID(bg_record.PlayerSpec)
                log("Found new spec: " .. tostring(name) .. tostring(description))
                table.insert(specs, id)
                seen_specs[id] = true
            end
        else
            log("Skipping empty spec.")
        end
    end
    return specs
end


WinningDataAccessLayer.get_record = function(self, record_idx)
    return self.ui_data[record_idx]
end

-- save bg data
WinningDataAccessLayer.save = function(self, bg_record)
    for k, v in pairs(WinningBattlegroundRecordModel) do
        assert(bg_record[k] ~= nil)
    end

    log("Saving BG record into DB.")

    -- TODO: Is this going to be slow eventually?
    -- Tested with 10k records and things were fine. Maybe things are fast
    -- enough not to worry about it yet.
    self.all_data[bg_record.uuid] = bg_record

    self:save_bg_data(bg_record)
end


WinningDataAccessLayer.convert_bg_record = function(self, bg_record)
    local ui_data = {}
    for i, v in pairs(self.addon.records_cols) do
        local val = bg_record[v['name']]
        if v['name'] == "Damage" then
            val = bg_record["DamageRaw"]
        elseif v['name'] == "Healing" then
            val = bg_record["HealingRaw"]
        elseif v['name'] == "Duration" then
            val = bg_record["DurationRaw"]
        elseif v['name'] == "Spec" then
            val = bg_record["PlayerSpec"]
        end
        ui_data[i] = val
    end
    return ui_data
end


-- Resync the data that is curated for the UI
WinningDataAccessLayer.resync = function(self, wiper)
    -- wipe the data stores for the UI
    wiper:wipe_ui_data()
    wiper:wipe_ui_bg_record_data()

    for i, bg_record in pairs(self.all_data) do
        log("Resyncing record: " .. tostring(i))
        self:save_bg_data(bg_record)
    end
end


WinningDataAccessLayer.save_bg_data = function(self, bg_record)
    if bg_record.uuid == nil then
        bg_record.uuid = Winning_UUID()
    end
    self.ui_bg_record_data[bg_record.uuid] = bg_record
    table.insert(self.ui_data, self:convert_bg_record(bg_record))
end


WinningDataAccessLayer.context_updated = function(self, ctx)
    log("Map context value: " .. tostring(ctx.map.value))
    local map_name, canEnter, isHoliday, isRandom, BattleGroundID, mapDescription = GetBattlegroundInfo(ctx.map.value)
    log("Map context real value: " .. tostring(map_name))

    log("Time context value: " .. tostring(ctx.time.value))

    if ctx.spec then
        log("Spec context value: " .. tostring(ctx.spec))
    end

    -- this will be slow :(
    for i, v in pairs(self.ui_data) do
        self.ui_data[i] = nil
    end

    for i, bg_record in pairs(self.all_data) do
        local map_match = false

        -- log(format("Checking Map Data...Id: %s, Map: %s", bg_record.MapId, bg_record.Map))
        -- log("Checking Map: " .. tostring(bg_record.Map))
        -- if ctx.map.value == 1 or ctx.map.value == bg_record.MapId then
        if ctx.map.value == 1 or map_name == bg_record.Map then
            map_match = true
        else
            --log(format("Map did not match: %s -> %s", map_name, bg_record.Map))
        end

        local time_match = false
        if ctx.time.context == self.ALL then
            time_match = true
        elseif ctx.time.context == self.SEASON then
            if bg_record.Season == ctx.time.value then
                time_match = true
            else
                log(format("TIME: Season did not match: %s", bg_record.Season))
            end
        elseif ctx.time.context == self.SIX_HOURS then
            -- log("Checking Six Hours time context.")
            -- log(format("%s >= %s", bg_record.Date, ctx.time.value))
            if bg_record.Date >= ctx.time.value then
                time_match = true
            else
                log(format("TIME: Six Hours did not match: %s", bg_record.Date))
            end
        else
            log("TIME: No context match for type? " .. ctx.time.context)
        end

        local type_match = false
        if ctx.type == self.ALL then
            type_match = true
            --log("Context type is All")
        else
            if ctx.type == self.CASUAL then
                if not bg_record.Rated then
                    type_match = true
                else
                    log("Context type is casual, BG was Rated.")
                end
            elseif ctx.type == self.RATED then
                if bg_record.Rated then
                    type_match = true
                else
                    log("Context type is Rated, BG was Casual.")
                end
            end
        end

        local spec_match = false
        if ctx.spec == 0 then
            spec_match = true
        else
            log("Checking player spec: " .. tostring(bg_record.PlayerSpec) .. " vs ctx spec: " .. tostring(ctx.spec))
            if bg_record.PlayerSpec == ctx.spec then
                spec_match = true
            end
        end

        if map_match and time_match and type_match and spec_match then
            self.ui_bg_record_data[bg_record.uuid] = bg_record
            table.insert(self.ui_data, self:convert_bg_record(bg_record))
        else
            self.ui_bg_record_data[bg_record.uuid] = nil
        end
    end

    self.addon:records_updated()
end

-- Calculate longest streak
WinningDataAccessLayer.longest_streak = function(self)
    local data = {}
    for uuid, bg_record in pairs(self.ui_bg_record_data) do
        table.insert(data, {date=bg_record.Date, record=bg_record})
    end
    table.sort(data, function(a, b)
        return a.date < b.date
    end)

    local streak_counter = 0
    local streak_record = 0
    local last_result_won = false
    for date, data_record in ipairs(data) do
        local bg_record = data_record.record
        if bg_record.Won then
            streak_counter = streak_counter + 1
            last_result_won = true
        else
            last_result_won = false
        end

        if not last_result_won then
            streak_counter = 0
        end

        if streak_counter > streak_record then
            streak_record = streak_counter
        end
    end
    return streak_record
end

-- Max of a particular statistic for a given context.
WinningDataAccessLayer.max_stat_in_ctx = function(self, stat_key)
    local val = 0
    for uuid, bg_record in pairs(self.ui_bg_record_data) do
        local new_val = bg_record[stat_key] or 0
        if new_val > val then
            val = new_val
        end
    end
    return val
end


-- Sum up all of a particular statistic for a given context.
WinningDataAccessLayer.summed_stat_in_ctx = function(self, stat_key)
    local val = 0
    for uuid, bg_record in pairs(self.ui_bg_record_data) do
        local new_val = bg_record[stat_key] or 0
        val = val + new_val
    end
    return val
end


-- Calculate wins and losses for current UI data.
WinningDataAccessLayer.win_loss = function(self)
    local wins = 0
    local losses = 0
    for uuid, bg_record in pairs(self.ui_bg_record_data) do
        if bg_record['Won'] then
            wins = wins + 1
        else
            losses = losses + 1
        end
    end
    return wins, losses
end


-- Calculate win/loss ratio for wins and losses.
WinningDataAccessLayer.win_loss_ratio = function(self, wins, losses)
    if wins == 0 then
        return "0%"
    end
    if losses == 0 and wins > 0 then
        return "100%"
    end
    if wins > 0 and losses > 0 then
        return format("%d%%", math.floor((wins/(wins+losses))*100))
    end
end


-- Calculate win/loss ratio for a specific map.
WinningDataAccessLayer.map_wins_losses = function(self, map)
    -- This currently does a string compare on the localized name of the map...
    -- but this is because there is no known way to get the current BG number
    -- for the BG that the player is in.
    --
    -- Further, since GetBattlegroundInfo and GetNumBattlegroundTypes both
    -- operate on battleground 'number' (id), this makes using the 'id' as an
    -- identifier a pain in the ass. I'd rather not create a hash of "Map ID"
    -- to battleground number.
    --
    -- TODO so slow.... make faster
    --
    local wins = 0
    local losses = 0

    for uuid, bg_record in pairs(self.all_data) do
        if bg_record.Map == map then
            if bg_record.Won then
                wins = wins + 1
            else
                losses = losses + 1
            end
        end
    end

    local data = {}
    data.wins = wins
    data.losses = losses
    return data
end


-- Calculate daily win ratio for a map and provide data appropriate for graphs.
WinningDataAccessLayer.daily_graph_data = function(self, map_name)
    log(format("Calculating daily graph data for: %s", map_name))

    local data = {}
    for i=1, 7 do
        data[i] = {}
        data[i].wins = 0
        data[i].losses = 0
    end

    local day_number
    for i, bg_record in pairs(self.all_data) do
        if bg_record.Map == map_name then
            day_number = tonumber(date("%w", bg_record.Date)) + 1
            log(format("Day number of: %s", day_number))
            if bg_record.Won then
                data[day_number].wins = data[day_number].wins + 1
            else
                data[day_number].losses = data[day_number].losses + 1
            end
        end
    end

    local bg_data = {}
    local win_data = {}
    local loss_data = {}
    local ratio
    for day, win_loss in pairs(data) do
        log(format("Checking day: %s, wins: %s, losses: %s", day, tostring(win_loss.wins), tostring(win_loss.losses)))
        if win_loss.wins == 0 and win_loss.losses == 0 then
            win_ratio = 0
        else
            win_ratio = math.ceil( (win_loss.wins / (win_loss.wins + win_loss.losses)) * 100)
        end
        if win_loss.losses > 0 then
            -- not really 100% loss, but we're filling in a stacked bar graph
            loss_ratio = 100
        else
            loss_ratio = 0
        end

        log(format("Win ratio: %s, Loss ratio: %s", win_ratio, loss_ratio))

        win_data[day] = {}
        win_data[day][1] = day
        win_data[day][2] = win_ratio

        loss_data[day] = {}
        loss_data[day][1] = day
        loss_data[day][2] = loss_ratio
    end

    bg_data.wins = win_data
    bg_data.losses = loss_data
    return bg_data
end
